from jabberwock.ccm.abstracts import BaseXType


class XPhoneLine(BaseXType):
    pass


class XCallForwardAll(BaseXType):
    pass


class XCallForwardAlternateParty(BaseXType):
    pass


class XCallForwardBusy(BaseXType):
    pass


class XCallForwardBusyInt(BaseXType):
    pass


class XCallForwardNoAnswer(BaseXType):
    pass


class XCallForwardNoAnswerInt(BaseXType):
    pass


class XCallForwardNoCoverage(BaseXType):
    pass


class XCallForwardNoCoverageInt(BaseXType):
    pass


class XCallForwardNotRegistered(BaseXType):
    pass


class XCallForwardNotRegisteredInt(BaseXType):
    pass


class XCallForwardOnFailure(BaseXType):
    pass


class XLineAppearanceAssociationForPresence(BaseXType):
    pass


class XLineAssociation(BaseXType):
    pass


class XEnduserMember(BaseXType):
    pass


class XPhone(BaseXType):
    pass



class AXLClientSettings(object):

    def __init__(self, host, user, password, path, version,
                 schema_path=None, suds_config=None, proxy=dict(),
                 transport_debugger=False):

        self.host = host
        self.user = user
        self.password = password
        self.path = path
        self.schema_path = schema_path
        self.suds_config = dict()
        self.proxy = proxy
        self.transport_debugger = transport_debugger
        if suds_config is not None:
            self.suds_config = suds_config
        self.version = '.'.join((str(version).split('.') + ['0'])[:2])


class ConfigurationRegistry(object):

    configurations = dict()

    def register(self, configuration, name='default'):
        self.configurations[name] = configuration

    def get(self, name='default'):
        return self.configurations[name]


registry = ConfigurationRegistry()



import logging
import types
from jabberwock.ccm.xtypes import XPhoneLine, XEnduserMember

log = logging.getLogger('pyaxl')


class MixingAbstractTemplate(object):
    """ Mixing class for all Types with template support.
    """

    @classmethod
    def template(cls, *args, typeclass=None, **kwargs):
        """ return with the given search criteria an complete object.
            On this object all attributes are pre-set with the
            values from template.
        """
        template = cls(*args, **kwargs)
        log.debug('%s created from template, criteria: %s, %s' % (cls.__name__, str(args), str(kwargs),))
        obj = template.clone()
        if typeclass is not None:
            obj['class'] = typeclass
        return obj


class MixingAbstractLines(object):

    def set_lines(self, lines):
        """Associate a list or a single ccm.Line object to device."""

        if not isinstance(lines, (list, types.GeneratorType)):
            lines = [lines]
        self.set_phone_lines(self._line_to_phone_line(lines))

    def set_phone_lines(self, phone_lines):
        """Associate a list or a single ccm.XPhoneLine object to device."""
        if not isinstance(phone_lines, (list, types.GeneratorType)):
            phone_lines = [phone_lines]
        self.lines = {'line': phone_lines}

    def remove_phone_lines(self, phone_lines):
        """Disassociate a list or a single ccm.XPhoneLine object from device."""
        if not isinstance(phone_lines, (list, types.GeneratorType)):
            phone_lines = [phone_lines]
        self.removeLines = {'line': phone_lines}

    def _line_to_phone_line(self, lines):
        """Convert a list or single ccm.Line object to XPhoneLine object."""
        if not isinstance(lines, (list, types.GeneratorType)):
            lines = [lines]
        for index, line in enumerate(lines):
            yield XPhoneLine(index=index + 1, dirn={'pattern': line.pattern, 'uuid': line.uuid})



class JabberwockException(Exception):
    pass


class UpdateException(JabberwockException):
    pass


class CreationException(JabberwockException):
    pass


class RemoveException(JabberwockException):
    pass


class ReloadException(JabberwockException):
    pass


class ResetException(JabberwockException):
    pass


class LogoutException(JabberwockException):
    pass


class NotAttachedException(JabberwockException):
    pass


class ProtocolException(JabberwockException):
    pass



import logging
from jabberwock import utils
from jabberwock.axlhandler import AXLClient

log = logging.getLogger('jabberwock')


class AXLSQL(object):

    def __init__(self, configname):
        self.client = AXLClient.get_client(configname)

    def _exec(self, sql):
        log.info('Execute SqlQuery "%s"' % sql)
        return self.client.axl.executeSQLQuery(sql)

    def _exec_update(self, sql):
        log.info('Execute SqlUpdate "%s"' % sql)
        return self.client.axl.executeSQLUpdate(sql)

    def _gen_result(self, dom_or_part, ispart=False):
        if not ispart:
            if 'row' not in dom_or_part['return']:
                return None
            li = dom_or_part['return']['row']
            if len(li) < 1:
                return None
            if len(li) > 1:
                raise ValueError('too many results.')
            dom_or_part = li[0]
        return {x.tag: x.text for x in dom_or_part}

    def _gen_result_list(self, dom):
        if not dom['return']:
            return None
        if 'row' not in dom['return']:
            return
        for part in dom['return']['row']:
            yield self._gen_result(part, True)

    def _to_bool(self, value):
        return 't' if bool(value) else 'f'


class AXLSQLUtils(AXLSQL):

    def user_phone_association(self, fkenduser):
        sql = 'SELECT * FROM enduserdevicemap WHERE fkenduser="%(fkenduser)s"'
        return self._gen_result_list(self._exec(sql % dict(fkenduser=utils.uuid(fkenduser))))

    def number_user_association(self, fknumplan):
        sql = 'SELECT * FROM endusernumplanmap WHERE fknumplan="%(fknumplan)s"'
        return self._gen_result_list(self._exec(sql % dict(fknumplan=utils.uuid(fknumplan))))

    def number_device_association(self, fknumplan):
        sql = ('SELECT d.name, d.pkid, n.dnorpattern AS dn FROM device AS d, '
               'numplan AS n, '
               'devicenumplanmap AS dnpm '
               'WHERE dnpm.fkdevice = d.pkid AND dnpm.fknumplan = n.pkid AND dnpm.fknumplan="%(fknumplan)s"')
        return self._gen_result_list(self._exec(sql % dict(fknumplan=utils.uuid(fknumplan))))

    def has_cups_cupc(self, fkenduser):
        sql = 'SELECT * FROM enduserlicense WHERE fkenduser="%(fkenduser)s"'
        return self._gen_result(self._exec(sql % dict(fkenduser=utils.uuid(fkenduser))))

    def insert_cups(self, fkenduser, cupc):
        sql = 'INSERT INTO enduserlicense (fkenduser, enablecups, enablecupc) VALUES ("%(fkenduser)s", "t", "%(cupc)s")'
        self._exec_update(sql % dict(fkenduser=utils.uuid(fkenduser), cupc=self._tobool(cupc)))

    def remove_cups(self, fkenduser):
        sql = 'DELETE FROM enduserlicense WHERE fkenduser = "%(fkenduser)s"'
        self._exec_update(sql % dict(fkenduser=utils.uuid(fkenduser)))

    def update_cups(self, fkenduser, cupc):
        sql = 'UPDATE enduserlicense SET enablecupc = "%(cupc)s" WHERE fkenduser = "%(fkenduser)s"'
        self._exec_update(sql % dict(fkenduser=utils.uuid(fkenduser), cupc=self._tobool(cupc)))

    def update_bfcp(self, fkenduser, bfcp):
        sql = 'UPDATE device SET enablebfcp = "%(bfcp)s" WHERE pkid = "%(fkenduser)s"'
        self._exec_update(sql % dict(fkenduser=utils.uuid(fkenduser), bfcp=self._tobool(bfcp)))

    def set_single_number_reach(self, fkremotedestination, value):
        sql = ('UPDATE remotedestinationdynamic SET enablesinglenumberreach = "%(value)s" '
               'WHERE fkremotedestination = "%(fkremotedestination)s"')
        self._exec_update(sql % dict(fkremotedestination=utils.uuid(fkremotedestination), value=self._tobool(value)))

    def get_single_number_reach(self, fkremotedestination):
        sql = ('SELECT enablesinglenumberreach FROM remotedestinationdynamic '
               'WHERE fkremotedestination = "%(fkremotedestination)s"')
        return self._gen_result(self._exec(sql % dict(fkremotedestination=utils.uuid(fkremotedestination))))

    def get_assigned_dn_list(self):
        sql = ("SELECT dnorpattern AS dn, MIN(r.name) AS name FROM numplan n, routepartition r "
               "WHERE r.pkid = n.fkroutepartition AND n.pkid IN (SELECT fknumplan FROM devicenumplanmap "
               "WHERE fkdevice IN (SELECT pkid FROM device)) GROUP BY dn ORDER BY dn")
        return self._gen_result_list(self._exec(sql))

    def get_inactive_dn_list(self):
        sql = ("SELECT n.pkid FROM numplan n LEFT OUTER JOIN devicenumplanmap m ON m.fkdevice = n.pkid "
               "WHERE m.fkdevice IS NULL AND n.tkpatternusage = '2' AND n.iscallable = 'f'")
        return self._gen_result_list(self._exec(sql))

    def get_users_with_self_service_id(self, self_service_id):
        sql = 'SELECT userid FROM enduser WHERE keypadenteredalternateidentifier LIKE "%(self_service_id)s"'
        return self._gen_result_list(self._exec(sql % dict(self_service_id='%' + self_service_id + '%')))

    def get_device_num_plan_map(self):
        sql = ("SELECT * FROM devicenumplanmap dnpm "
               "WHERE dnpm.fknumplan IN (SELECT n.pkid "
               "FROM numplan AS n INNER JOIN routepartition AS rp ON n.fkroutepartition=rp.pkid "
               "WHERE rp.name IN "
               "('SG-AA-Internal', 'SG-DA-Internal', 'SG-DT-Internal', 'SG-PH-Internal', 'SG-SH-Internal'))")
        return self._gen_result_list(self._exec(sql))


from jabberwock.ccm.abstracts import BaseCUCMModel
from jabberwock.ccm.mixings import MixingAbstractLines
from jabberwock.ccm.mixings import MixingAbstractTemplate
from jabberwock.axlsql import AXLSQLUtils
from jabberwock import exceptions
from jabberwock import utils


class DeviceProfile(BaseCUCMModel,
                    MixingAbstractTemplate,
                    MixingAbstractLines):
    pass
    
    
class User(BaseCUCMModel, MixingAbstractTemplate):

    def add_associated_devices(self, phones):
        if not self.associatedDevices:
            self.set_associated_devices(phones)
            return
        if not isinstance(phones, list):
            phones = [phones]
        if not isinstance(phones[0], str):
            phones = [i.name for i in phones]
        union = list(set(phones) | set(self.associatedDevices['device']))
        self.associatedDevices = dict(device=union) if union else ''

    def set_associated_devices(self, phones):
        if not isinstance(phones, list):
            phones = [phones]
        if not isinstance(phones[0], str):
            phones = [i.name for i in phones]
        self.associatedDevices = dict(device=phones) if phones else ''

    def set_cti_controlled_device_profiles(self, device_profiles):
        if not isinstance(device_profiles, list):
            device_profiles = [device_profiles]
        self.ctiControlledDeviceProfiles = dict(profileName=[dict(uuid=i.uuid) for i in device_profiles]) \
            if device_profiles else ''

    def set_phone_profiles(self, device_profiles):
        if not isinstance(device_profiles, list):
            device_profiles = [device_profiles]
        self.phoneProfiles = dict(profileName=[dict(uuid=i.uuid) for i in device_profiles]) if device_profiles else ''

    def get_mobility_association(self):
        """
        Return phones that are associated with this user.
        """
        sql_utils = AXLSQLUtils(self.__config_name__)
        if not self.__attached__:
            raise exceptions.NotAttachedException('User is not attached')
        for i in sql_utils.user_phone_association(self.uuid):
            yield Phone(uuid=i['fkdevice'])

    def get_cups_cupc(self):
        cups, cupc, pkid = self._get_cups_cupc()
        return cups, cupc

    def set_cups_cupc(self, cups, cupc):
        sqlutils = AXLSQLUtils(self.__config_name__)
        if cupc and not cups:
            raise exceptions.PyAXLException('If cupc is true, cups must also be true')
        rcups, rcupc, pkid = self._get_cups_cupc()
        if rcups is None and cups:
            sqlutils.insert_cups(self._uuid, cupc)
        elif rcups and not cups:
            sqlutils.remove_cups(self._uuid)
        elif rcups and cups:
            sqlutils.update_cups(self._uuid, cupc)

    def _get_cups_cupc(self):
        sql_utils = AXLSQLUtils(self.__config_name__)
        if not self.__attached__:
            raise exceptions.NotAttachedException('User is not attached')
        re = sql_utils.has_cups_cupc(self.uuid)
        if re is None:
            return None, None, None
        return re['enablecups'] == 't', re['enablecupc'] == 't', re['pkid']


class UserGroup(BaseCUCMModel):
    pass


class Line(BaseCUCMModel):
    def __init__(self, **kwargs):
        kwargs.setdefault('usage', 'Device')
        super().__init__(**kwargs)

    def get_primary_users(self):
        """ Return users that have this line set as a primary extension."""
        sql_utils = AXLSQLUtils(self.__config_name__)
        if not self.__attached__:
            raise exceptions.NotAttachedException('Line is not attached')
        for i in sql_utils.number_user_association(self.uuid):
            yield User(uuid=i['fkenduser'])

    def get_associated_devices(self):
        """ Return devices that are associated to this directory number."""
        sql_utils = AXLSQLUtils(self.__config_name__)
        if not self.__attached__:
            raise exceptions.NotAttachedException('Line is not attached')
        for i in sql_utils.number_user_association(self.uuid):
            yield User(uuid=i['fkenduser'])


class TransPattern(BaseCUCMModel):
    pass

    
class Phone(BaseCUCMModel,
            MixingAbstractTemplate,
            MixingAbstractLines):

    def get_vendor_config(self):
        return utils.elements_to_dict(self.vendorConfig['_value_1'])

    def set_vendor_config(self, **kwargs):
        self.vendorConfig = dict(_value_1=utils.dict_to_elements(kwargs))
            
    def logout(self):
        if not self.__attached__:
            raise exceptions.LogoutException('Phone is not attached')
        self.__client__.service.doDeviceLogout(dict(_uuid=self._uuid))

    def login(self, user, deviceProfile, duration=1):
        if not self.__attached__:
            raise exceptions.LogoutException('Phone is not attached')
        self.__client__.service.doDeviceLogin(dict(_uuid=self._uuid),
                                              duration,
                                              dict(_uuid=deviceProfile._uuid),
                                              user.userid)

            
class AppUser(BaseCUCMModel):
    pass


class CallPickupGroup(BaseCUCMModel):
    pass


class Css(BaseCUCMModel):
    pass


class CtiRoutingPoint(BaseCUCMModel):
    pass  

    
class DevicePool(BaseCUCMModel):
    pass


class Gateway(BaseCUCMModel):
    pass


class GatewayEndpointAnalogAccess(BaseCUCMModel):
    pass


class HuntList(BaseCUCMModel):
    pass


class HuntPilot(BaseCUCMModel):
    pass


class LineGroup(BaseCUCMModel):
    pass


class PhoneButtonTemplate(BaseCUCMModel):
    pass


class RoutePartition(BaseCUCMModel):
    pass


class RouteList(BaseCUCMModel):
    pass


class VoiceMailPilot(BaseCUCMModel):
    pass


class VoiceMailProfile(BaseCUCMModel):
    pass


class RemoteDestination(BaseCUCMModel):

    def set_single_number_reach(self, value):
        """ Set single number reach flag is not possible in version 10.5
            see ticket: https://supportforums.cisco.com/discussion/12438721/single-number-reach-axl
            The workaround is again to set some value with SQL! Yupiiiii!
        """
        if not self.__attached__:
            raise exceptions.NotAttachedException('User is not attached')
        sqlutils = AXLSQLUtils(self.__config_name__)
        sqlutils.set_single_number_reach(self._uuid, value)

    def get_single_number_reach(self):
        if not self.__attached__:
            raise exceptions.NotAttachedException('User is not attached')
        sqlutils = AXLSQLUtils(self.__config_name__)
        value = sqlutils.get_single_number_reach(self._uuid)
        return value['enablesinglenumberreach'] == 't'


class RemoteDestinationProfile(BaseCUCMModel,
                               MixingAbstractLines,
                               MixingAbstractTemplate):
    @classmethod
    def template(cls, *args, **kwargs):
        return super().template(*args, typeclass='Remote Destination Profile', **kwargs)


class TodAccess(BaseCUCMModel):
    pass


class TimeSchedule(BaseCUCMModel):

    def removeMembers(self, members):
        if not isinstance(members, list):
            members = [members]
        removeMembers = [dict(member=dict(timePeriodName=dict(_uuid=uuid))) for uuid in members]
        self.__client__.service.updateTimeSchedule(removeMembers=removeMembers, uuid=self._uuid)

    def addMembers(self, members):
        if not isinstance(members, list):
            members = [members]
        addMembers = [dict(member=dict(timePeriodName=dict(_uuid=uuid))) for uuid in members]
        self.__client__.service.updateTimeSchedule(addMembers=addMembers, uuid=self._uuid)


class TimePeriod(BaseCUCMModel):
    pass


def main():
    return


if __name__ == "__main__":
    main()



from zeep import Client
from zeep.transports import Transport
from zeep.plugins import HistoryPlugin
from requests import Session
from requests.auth import HTTPBasicAuth
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning
from zeep.cache import SqliteCache
from zeep.settings import Settings
import configparser


class AXLClient(Client):
    """
    The AXLClient class sets up the connection to the call manager with methods for configuring UCM
    Tested with environment of Python 3 and zeep.
    """

    clients = dict()
    BINDING_NAME = "{http://www.cisco.com/AXLAPIService/}AXLAPIBinding"

    def __init__(self, config_name='default', toolkit_path='/axlsqltoolkit', **kwargs):
        """
        Args:
            config_name: the name of the config file
        """
        self.config = configparser.ConfigParser()
        self.config.read('{toolkit_path}/config/{config_name}.ini'.format(toolkit_path=toolkit_path,
                                                                          config_name=config_name))
        username = self.config['authentication']['username']
        password = self.config['authentication']['password']
        version = self.config['authentication']['version']
        disable_warnings(InsecureRequestWarning)
        settings = Settings(strict=False, xml_huge_tree=True)
        wsdl = 'file://' + toolkit_path + '/schema/' + version + '/AXLAPI.wsdl'
        session = Session()
        session.verify = False
        session.auth = HTTPBasicAuth(username, password)
        transport = Transport(cache=SqliteCache(), session=session, timeout=20)
        defaults = dict(wsdl=wsdl, transport=transport, plugins=[HistoryPlugin()], settings=settings)
        merged_kwargs = {**defaults, **kwargs}
        super().__init__(**merged_kwargs)

    @property
    def axl(self):
        server = self.config['authentication']['server']
        address = "https://{server}:8443/axl/".format(server=server)
        return self.create_service(self.BINDING_NAME, address)

    @property
    def factory(self):
        return self.type_factory('ns0')

    @classmethod
    def get_client(cls, config_name='default', toolkit_path='/axlsqltoolkit', recreate=False):
        """ return a single instance of client for each configuration.
        """
        client = None
        if config_name not in cls.clients or recreate:
            client = AXLClient(config_name, toolkit_path)
        return cls.clients.setdefault(config_name, client)


def main():
    return


if __name__ == "__main__":
    main()



import re
from lxml.builder import E

REGEX_UUID = re.compile(r'^\{?([0-9a-f]{8}-(?:[0-9a-f]{4}-){3}[0-9a-f]{12})\}?$')


def uuid(value):
    """ parse uuid and return a blank uuid without "{}".
        if the uuid is wrong it will raise an exception.
    """
    re = REGEX_UUID.match(value.lower())
    if re is None:
        raise ValueError('uuid is wrong')
    returnvalue, = re.groups()
    return returnvalue


def axlbool(value):
    """ convert text to python bool
    """
    if value is None:
        return None
    if not value:
        return False
    if value.lower() == 'true':
        return True
    return False


def dict_to_elements(dict):
    return [E(k, v) for k, v in dict.items()]


def elements_to_dict(elements):
    return {e.tag: e.text for e in elements}



from setuptools import setup, find_packages
from dunamai import Version

setup(
    name="jabberwock",
    version=Version.from_any_vcs().base,
    author="Kyle Wroble",
    author_email="kwroble@gmail.com",
    url="https://github.com/kwroble/jabberwock",
    description="Python library that helps with accessing Cisco Call Manager over the AXL interface",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Zope Public License",
        "Operating System :: OS Independent",
    ],
    install_requires=["appdirs", "attrs", "boltons", "cached-property", "certifi", "chardet", "defusedxml", "idna",
                      "isodate", "lxml", "pytz", "requests", "requests-toolbelt", "six", "urllib3", "zeep", "dunamai"],
)



import logging
from boltons.iterutils import remap
from zeep.xsd.valueobjects import CompoundValue
from jabberwock.axlhandler import AXLClient
from jabberwock import exceptions


PF_LIST = 'list'
PF_GET = 'get'
PF_UPDATE = 'update'
PF_ADD = 'add'
PF_REMOVE = 'remove'
PF_RESET = 'reset'
XSD_NS = 'ns0'

log = logging.getLogger('jabberwock')


class BaseCUCMModel(object):
    """
    Provide base functionality for all logical entities in Cisco Unified Communications Solutions (CUCM).

    This will make the bridge between zeep and CUCM objects.
    In addition, all standard AXL operations are implemented here.

    Attributes:
        __name__: Name of the logical CUCM object (e.g. User, Phone, RoutePartition)
        __config__: Name of the configuration
        __client: Client object used to communicate with AXL API
        __attached__: Is this object associated with an AXL object?
        __updateable__: List containing all changed attributes since last object load
    """

    __config_name__ = ''
    __config__ = None
    __client__ = None
    __attached__ = False
    __update_request__ = None
    __update_substitutions__ = None
    uuid = None

    def __init__(self, *args, **kwargs):
        """
        Create a new CUCM object.
        """
        config_name = kwargs.pop('config_name', 'default')
        self._configure(config_name)
        self._initialize(**kwargs)

    def __setattr__(self, name, value):
        """
        Set the value of an attribute.

        Keeps track of the name of the attribute that is changed. Generally used for the update method.
        """
        builtin = name.startswith('__') and name.endswith('__')
        if not builtin:
            if self.__attached__:
                self.__setattr_update__(name, value)
        super().__setattr__(name, value)

    def __setattr_update__(self, name, value):
        if name in self.__update_substitutions__.keys():
            name = self.__update_substitutions__[name]
        setattr(self.__update_request__, name, value)

    @classmethod
    def _axl_operation(cls, prefix, name, client):
        """
        Return an AXL operation.
        """
        return getattr(client.axl, '%s%s' % (prefix, name,))

    @classmethod
    def _prepare_result(cls, result, returns):
        """

        :param result:
        :param returns:
        :return:
        """
        unwrapped = result['return']
        if not unwrapped:
            return
        if isinstance(unwrapped, str):
            return
        unwrapped = getattr(unwrapped, cls._first_lower(cls.__name__))
        for obj in unwrapped:
            yield {key: value for (key, value) in obj.__dict__['__values__'].items() if key in returns}

    def _initialize(self, **kwargs):
        """
        a part of init method. If some search criteria was found it
            will automatically load this object.
        """
        self._load(**kwargs)
        self.__update_request__ = self._get_update_request()
        self.__update_substitutions__ = self._get_update_substitutions(self.__update_request__)
        if self.uuid:
            self.__attached__ = True

    def _load(self, **kwargs):
        """
        Get the specified object from Call Manager and merge its attributes with this CUCM object.
        """
        operation = self._axl_operation(PF_GET, self.__name__, self.__client__)
        uuid = kwargs.pop('uuid', '')
        request = getattr(self.__client__.factory, '%s%sReq' % (PF_GET.capitalize(), self.__name__))()
        if uuid:
            criteria = {'uuid': uuid}
        else:
            criteria = {key: value for (key, value) in kwargs.items() if key in request.__dict__['__values__'].keys()}
        try:
            result = operation(**criteria)
            result = getattr(getattr(result, 'return'), self._first_lower(self.__name__))
        except:
            print('Unable to get object. Creating xtype...')
            result = self._get_xtype(**kwargs)
        self._loadattr(result)

    @classmethod
    def _first_lower(cls, name):
        return name[:1].lower() + name[1:] if name else ''

    @classmethod
    def _strip_empty_tags(cls, obj):
        """
        Recursively strip all attributes equal to -1 from an object.

        The AXL update operation can't handle values of -1. This will recursively create a copy of object and remove
        all attributes with a value of -1.
        """
        def visit(path, key, value):
            if value == -1:
                return False
            elif isinstance(value, CompoundValue):
                for i in dir(value):
                    if getattr(value, i) == -1:
                        return False
            return key, value

        return remap(obj, visit=visit)

    @property
    def __name__(self):
        return self.__class__.__name__

    def _configure(self, config_name):
        """
        Set up AXL client connection.
        """
        self.__client__ = AXLClient.get_client(config_name)
        self.__config_name__ = config_name

    def _loadattr(self, obj):
        """
        Copy all attributes from an AXL object to this CUCM object.
        """
        for k, v in obj.__dict__['__values__'].items():
            self.__setattr__(k, v)

    def _get_xtype(self, **kwargs):
        """
        Return an XType AXL object.
        """
        x_type = getattr(self.__client__.factory, 'X%s' % self.__name__)()
        tags = dir(x_type)
        kwargs = {key: value for (key, value) in kwargs.items() if key in tags}
        kwargs = self._strip_empty_tags(kwargs)
        x_type = getattr(self.__client__.factory, 'X%s' % self.__name__)(**kwargs)
        return x_type

    def _get_update_request(self, **kwargs):
        return getattr(self.__client__.factory, '%s%sReq' % (PF_UPDATE.capitalize(), self.__name__))(**kwargs)

    def _get_update_substitutions(self, update_request):
        subs = {}
        for key in dir(update_request):
            if key.startswith('new'):
                tag = self._first_lower(key[3:])
                if tag in dir(self):
                    subs[tag] = key
        return subs

    def create(self):
        """
        Add this object to CUCM.
        """
        if self.__attached__:
            raise exceptions.CreationException('this object is already attached')
        operation = self._axl_operation(PF_ADD, self.__name__, self.__client__)
        x_type = getattr(self.__client__.factory, 'X%s' % self.__name__)()
        tags = dir(x_type)
        unwrapped = {key: value for (key, value) in self.__dict__.items() if key in tags}
        x_type = self._get_xtype(**unwrapped)
        result = operation(x_type)
        self.uuid = result['return']
        self.__attached__ = True
        self.__update_request__ = self._get_update_request(uuid=self.uuid)
        log.info('new %s was created, uuid=%s' % (self.__name__, self.uuid,))
        return self.uuid

    def update(self):
        """
        Update the CUCM object with all changes made to this object.
        """
        if not self.__attached__:
            raise exceptions.UpdateException('you must create an object with "create" before update')
        self.__update_request__.uuid = self.uuid
        operation = self._axl_operation(PF_UPDATE, self.__name__, self.__client__)
        operation(**self.__update_request__.__dict__['__values__'])
        self.__update_request__ = self._get_update_request(uuid=self.uuid)
        log.info('%s was updated, uuid=%s' % (self.__name__, self.uuid,))

    def remove(self):
        """
        Delete this object from CUCM.
        """
        if not self.__attached__:
            msg = 'This object is not attached and can not removed from CUCM'
            raise exceptions.RemoveException(msg)
        operation = self._axl_operation(PF_REMOVE, self.__name__, self.__client__)
        operation(uuid=self.uuid)
        self.uuid = None
        self.__attached__ = False
        self.__update_request__ = self._get_update_request()
        log.info('%s was removed, uuid=%s' % (self.__name__, self.uuid,))

    def reload(self):
        """
        Reload this object from CUCM.
        """
        if not self.__attached__:
            msg = 'This object is not attached and can not be reloaded from CUCM'
            raise exceptions.ReloadException(msg)
        self._load(uuid=self.uuid)
        self.__update_request__ = self._get_update_request(uuid=self.uuid)

    def reset(self):
        """
        Reset this object.
        """
        if not self.__attached__:
            msg = 'This object is not attached and can not be reset.'
            raise exceptions.ResetException(msg)
        operation = self._axl_operation(PF_RESET, self.__name__, self.__client__)
        try:
            operation(uuid=self.uuid)
        except:
            print("Unable to reset object.")

    def clone(self):
        """
        Return a clone of this object.

        After cloning, the new object will be detached.
        This means the cloned object can be directly added to CUCM with the create function.
        """
        obj = self.__class__()
        obj.__dict__.update(self.__dict__)
        obj.uuid = None
        obj.__attached__ = False
        log.debug('%s was cloned' % self.__name__)
        return obj

    @classmethod
    def list(cls, criteria, returns, skip=None, first=None, configname='default'):
        """

        :param criteria: Dictionary of search criteria.
        :param returns: List of the desired returned tags.
        :param skip: The number of results to skip, starting at the first.
        :param first: The maximum number of results to return, starting at the first.
        :param configname: Name of the configuration. Default value is 'default'
        :yeild: Returns the matching search results.
        """
        client = AXLClient.get_client(configname)
        operation = cls._axl_operation(PF_LIST, cls.__name__, client)
        tags = dict([(i, '') for i in returns])
        log.debug('fetch list of %ss, search criteria=%s' % (cls.__name__, str(criteria)))
        args = criteria, tags
        if skip is not None or first is not None:
            if skip is None:
                skip = 0
            if first is None:
                args = criteria, tags, skip
            else:
                args = criteria, tags, skip, first
        return cls._prepare_result(operation(*args), returns)

    @classmethod
    def list_obj(cls, criteria, skip=None, first=None, configname='default'):
        """
        Return all objects that match the given search criteria.

        The return value is generator. Each next() call will fetch a new instance and return it as an object.
        """
        for obj in cls.list(criteria, 'uuid', skip, first, configname):
            yield cls(uuid=obj['uuid'])


class BaseXType(object):

    def __new__(cls, *args, **kwargs):
        client = AXLClient.get_client()
        return getattr(client.factory, cls.__name__)(*args, **kwargs)


class BaseXTypeListItem(dict):
    """ A special XType that can be used to fill into a list.
    """
    def __init__(self, *args, **kwargs):
        name = self.__class__.__name__
        xtype = type(name, (BaseXType, self.__class__), dict())(*args, **kwargs)
        self[name[1:]] = xtype
